#!perl -w
# $Id: $
# ChangeBootDriveLetter - Scan volume and edit files for boot drive letter change
# Use -f to force work file rebuild
# Uses/builds filename work file in %TMP if list file omitted

# Copyright (c) 2009 Steven Levine and Associates, Inc.
# All rights reserved.

# This program is free software licensed under the terms of the GNU
# General Public License.  The GPL Software License can be found in
# gnugpl2.txt or at http://www.gnu.org/licenses/licenses.html#GPL

# 13 Aug 09 SHL Baseline

#== TBD ==
# fixme to normalize path slashes unix_slash dos_slash
# fixme to track skipped files for restarts
#== end TBD ==

use strict;
use warnings;

use File::Spec;
use File::Spec::Functions;
# 14 Aug 09 SHL	fixme to use
# use File::stat;

use subs qw( ask_action gobble_file quote_for_system quote_for_shell usage );

use File::Basename;
my $cmd_name;
my $tag_name;
get_cmd_name($0);

my $shell = $ENV{'OS2_SHELL'};
$shell =~ s/\\/\//g;

use Getopt::Std;
my %opts;
getopts('fhltV', \%opts) or exit;

print 'Version 0.1' and exit if $opts{V};

# Derive file names from script file base name (i.e. tag_name)
my $def_wrkfile = File::Spec->catfile($ENV{'TMP'}, "$tag_name.lst");
my $def_iniwrkfile = File::Spec->catfile($ENV{'TMP'}, "$tag_name-ini.lst");
my $logfile = File::Spec->catfile($ENV{'TMP'}, "$tag_name.log");

my $usage_msg = "Usage: $cmd_name [-f] [-h] [-l] [-t] [-V] [list-file] [old-drv [new-drv]]";

print <<EOD and exit if $opts{h};
Scan file name list and edit files as needed
Generates listing if not found or if forced

$usage_msg

 -f          Refresh status listing before checking
 -h          Display this message
 -l          Generate list of files selected for editing
 -t          Run in test mode updates written to filename~
 -V          Display version

 list-file   dir /f list of file to process, defaults to $def_wrkfile
 old-file    original drive letter
 new-file    new drive letter

 List files generated by -l are named $tag_name-xxx.lst where xxx is unique
 Files are written to TMP directory
 Originals are backed up to filename~
EOD

my ($wrkfile, $iniwrkfile, $old_drv, $new_drv, $s, $rc, $cmd);

while (@ARGV) {
  $s = shift @ARGV;
  $s = "$s:" if $s =~ /^[[:alpha:]]$/;	# Add :
  if ($s =~ /^[[:alpha:]]:$/) {
    if (!defined $old_drv) {
      $old_drv = $s;
    } elsif (!defined $new_drv) {
      $new_drv = $s;
    } else {
      usage 'Only two drive letter specs allowed';
    }
  } elsif (!defined $wrkfile) {
    usage "$s not found" if ! -f $s;
    $wrkfile = $s;
  } elsif (!defined $iniwrkfile) {
    usage "$s not found" if ! -f $s;
    $iniwrkfile = $s;
  } else {
    usage 'Already have two work files';
  }
} # while

# Set defaults

$wrkfile = $def_wrkfile if !defined $wrkfile;
$iniwrkfile = $def_iniwrkfile if !defined $iniwrkfile;
# $old_drv does not need to exist - used only to find refs to old path
$old_drv = 'F:' if !defined $old_drv;
# $new_drv must exist or bad things will happen
if (!defined $new_drv) {
  $new_drv = $ENV{'HOSTNAME'} eq 'slat42-1' ? 'L:' : 'K';
}

$old_drv =~ tr/a-z/A-Z/;			# Ensure upper case
my $old_drvl = $old_drv;
$old_drvl =~ tr/A-Z/a-z/;			# Ensure case version

$new_drv =~ tr/a-z/A-Z/;
my $new_drvl = $new_drv;
$new_drvl =~ tr/A-Z/a-z/;

my $refreshwrkfile = defined($opts{f}) || ! -f $wrkfile || -s $wrkfile < 8;
my $refreshiniwrkfile = defined($opts{f}) || ! -f $iniwrkfile || -s $iniwrkfile < 8;

$| = 1;			# Force flush on stdout

open LOG, ">>$logfile" or die "open $logfile $!";
warn "Opened logfile $logfile\n";

$_ = "$cmd_name invoked at " . localtime() . "\n";
warn $_;
print LOG "\n", $_;
$_ = "Using $wrkfile as basis to change $old_drv to $new_drv\n";
warn $_;
print LOG $_;

my $listfile;

if ($opts{l}) {
  # Select unique list file name
  my $seq = -1;
  while (1) {
    $seq++;
    $seq = substr("00$seq", -3);
    $listfile = File::Spec->catfile($ENV{'TMP'}, "$tag_name-$seq.lst");
    last if ! -f $listfile;
  }
  open LIST, ">$listfile" or die "open $listfile $!";
  warn "Opened list file $listfile\n";
}

# $|=0;

my $basenames_to_skip = ' license readme readme2 ';
my $fullnames_to_skip = 'blistlay.out dir doskrnl gradd license os2krnl os2boot os2dump os2ldr readme readme2 readme.txt vga vgados viotbl.dcp viotbl.iso';
my $exts_to_skip = ' add au avi bak bas bmp com class dll dmp doc drv dsp dtd' .
		   ' eng exe flt fnt fon gif gm' .
		   ' jar jpg' .
		   ' h hlp htm html ico ifs img inf info ipf itr' .
		   ' java js lib l1 l2 log lut map mapx mid mo msg mov mpg mpeg' .
		   ' os2 pdf pf pfb pfm png psd ptr sbk sdf swf sym tdf tmf ttf txt wa0 wav wpd wpi zip ';

if ($refreshwrkfile) {
  warn "Generating $wrkfile with dir. Please wait.\n";
  my $rc = system("$shell /c dir /a:-d /f /s $new_drv\\ >$wrkfile");
  # echo error 2 sometimes
  $rc = 0 if $rc == 2 << 8;
  die "dir /f failed rc = $rc $!(", 0+$!, ")\n" if $rc;
  warn "$wrkfile generated\n";
}

if ($refreshiniwrkfile) {
  warn "Generating $iniwrkfile with dir \*.ini. Please wait.\n";
  my $rc = system("$shell /c dir /a /f /s $new_drv\\\*.ini >$iniwrkfile");
  $rc = 0 if $rc == 2 << 8;
  die "dir /f failed rc $rc $!\n" if $rc;
  warn "$iniwrkfile generated\n";
}

warn "Processing $wrkfile\n";

my $file_cnt = 0;
open(IN, "<$wrkfile");
while(<IN>) {
  $file_cnt++;
}
close IN;

open(IN, "<$wrkfile");

my $files_matched = 0;
my $files_changed = 0;
my $drives_changed = 0;

while(<IN>) {

  chomp;
  my $cur_file = $_;
  my $file_num = $.;

  next if -d $cur_file;

  # warn $cur_file;

  my (@a, $ext);
  @a = File::Spec->splitpath($cur_file);
  $a[2] =~ tr/A-Z/a-z/;		# Lowercase name
  next if index($fullnames_to_skip, " $a[2] ") != -1;	# Skip based on full name

  if ($a[2] =~ /^(.+)\.(.*)$/) {
    next if index($basenames_to_skip, " $1 ") ne -1;	# Skip based on basename
    next if index($exts_to_skip, " $2 ") ne -1;		# Skip based on extension
  }

  my $bkup_file = "$cur_file~";

  # If testing reference backup if it exists
  if ($opts{t} && -f $bkup_file) {
    $cur_file = $bkup_file
  }

  my $buf = gobble_file($cur_file);

  die if !defined $buf;

  my $buf_len = length($buf);
  next if !$buf_len;		# File empty

  my $is_binary = ($buf =~ /([^[:print:]\t\n\r\x1a])/);

  if ($is_binary) {
    # Check for executables
    next if substr($buf, 0, 2) eq 'MZ';
    next if substr($buf, 0, 2) eq 'LX';
  }

  my $offsetu = 0;
  my $offsetl = 0;
  my $offset;
  my $repl;
  my $is_lower;
  my $changes = 0;
  my $auto_changes = 0;
  my $first = 1;
  my $yes_to_all = 0;
  my $path;
  my %known_actions;			# Indexed by path
  undef %known_actions;
  my $undo_buf = $buf;
  while (1) {
    $offsetu = index($buf, $old_drv, $offsetu) if $offsetu != -1;
    $offsetl = index($buf, $old_drvl, $offsetl) if $offsetl != -1;
    # 15 Aug 09 SHL fixme to prompt if any auto changes
    if ($offsetu == -1 && $offsetl == -1) {
      if ($auto_changes) {
      }
      last;
    }
    if ($offsetu == -1 || ($offsetl != -1 && $offsetl <= $offsetu)) {
      $offset = $offsetl;
      $is_lower = 1;
      $repl = $new_drvl;
    }
    else {
      $offset = $offsetu;
      $is_lower = 0;
      $repl = $new_drv;
    }

    # Try to extract full path
    $s = substr($buf, $offset, 260);
    $s =~ /^([[:alpha:]]:[-[:alnum:]_.\$\\\/]*)/;	# Trim non-filename characters
    my $path = $1;
    $s = $path;
    while (!-d $path && !-f $path) {
      chop $path;
      $path =~ s/\\$//;
      last if length($path) <= 2;
    } # while chopping
    if (length($path) > 3) {
      $s = $path;
    } else {
      $path = $s;
      if (length($path) == 2 ||	substr($path, 2, 1) ne "\\") {
	if (!$is_binary) {
	  # Show line containing drive reference
	  my $i = $offset;
	  while ($i > 0 and substr($buf, $i - 1, 1) =~ /[^\n\r]/s) {
	    $i--;
	  }
	  $s = substr($buf, $i, $offset - $i + 40);
	  $s =~ s/[\r\n].*//s;
	}
	else {
	  # Show data surrounding path reference
	  $s = substr($buf, $offset >= 35 ? $offset - 35 : $offset, 70);
	  $s =~ s/[^[:print:]]/ /g;
	}
      }
    }

    if ($first) {
      $first = 0;
      warn "\n";
      $files_matched++;
    }

    my $t = $is_binary ? 'binary' : 'text';
    if ($path ne $s) {
      # warn "Found string $s at offset $offset/$buf_len of $t file $cur_file ($file_num)\n";
      warn "$cur_file ($t file #$file_num/$file_cnt)\nString $s at offset $offset/$buf_len\n";
    } else {
      # warn "Found path $path at offset $offset/$buf_len of $t file $cur_file ($file_num)\n";
      warn "$cur_file ($t file #$file_num/$file_cnt)\nPath $path at offset $offset/$buf_len\n";
      warn "Path does not exist\n" if !-d $path && !-f $path;
    }

    if ($opts{l}) {
      printf LIST "$cur_file\n";	# Output to list file for subset list
      last;
    }

    my $action = undef;
    if ($yes_to_all >= 2) {
      $action = 'y';
    }
    if ($yes_to_all >= 1 && $path =~ /^..\\/) {
      $action = 'y';
    }
    elsif (exists $known_actions{$path}) {
      $action = $known_actions{$path};
    }

    if (!defined $action) {
      $action = ask_action;
    }
    if ($action eq 'a') {
      $yes_to_all = 1;			# Paths with \
      next if $path !~ /^..\\/;		# Ask again if no \
      $action = 'y';
      $auto_changes++;
    }
    if ($action eq 'A') {
      $yes_to_all = 2;			# Any path
      $action = 'y';
      $auto_changes++;
    }

    if ($action eq 'n') {
    }
    elsif ($action eq 'N') {
      # 14 Aug 09 SHL fixme to complain
      $known_actions{$path} = $action if length($path) > 2;
    }
    elsif ($action eq 'y') {
      substr($buf, $offset, 2) = $repl;
      print LOG "Changed $path at offset $offset to reference $repl in $cur_file\n";
      $changes++;
    }
    elsif ($action eq 'Y') {
      $known_actions{$path} = $action if length($path) > 2;
      substr($buf, $offset, 2) = $repl;
      print LOG "Changed $path at offset $offset to reference $repl in $cur_file\n";
      $changes++;
    }
    elsif ($action eq 'r') {
      $offsetu = 0;
      $offsetl = 0;
      $buf = $undo_buf;
      $changes = 0;
      next;
    }
    elsif ($action eq 'R') {
      $changes = 0;
      last;
    }
    elsif ($action eq 's') {
      last;
    }
    elsif ($action eq 'q') {
      exit;
    }
    else {
      die "action $action unexpected\n";
    }

    if ($is_lower) {
      $offsetl++;
    } else {
      $offsetu++;
    }
  } # while offset

  if ($changes) {
    warn "\n";
    # Will be same file if first time test
    if (!-f $bkup_file && $cur_file ne $bkup_file) {
      warn "Backing up $cur_file to $bkup_file\n";
      $s = quote_for_shell(quote_for_system $cur_file);
      my $t = quote_for_shell(quote_for_system $bkup_file);
      $cmd = "$shell /c copy /q /h $s $t";
      warn "$cmd\n";
      # system($cmd) or die "copy failed rc = $!(", 0+$!, ")\n";
      my $rc = system($cmd);
      die "copy failed rc = $rc $!(", 0+$!, ")\n" if $rc;
    }
    # If test mode apply changes to backup rather than original
    $cur_file = $bkup_file if $opts{t};
    warn "Saving $changes changes to $cur_file\n";
    # Get current attributes/time
    # 14 Aug 09 SHL fixme to know why this fails
    die "$cur_file is gone" if ! -f $cur_file;
    my ($dev,$ino,$mode,$nlink,$uid,$gid,$rdev,$size,
	$atime,$mtime,$ctime,$blksize,$blocks) = stat($cur_file)
      or die "stat $cur_file $!\n";
    die "stat $cur_file failed mode\n" if !defined $mode;
    warn "Rewriting $cur_file\n";
    chmod 0777, $cur_file or die "chmod failed $!\n";
    $s = quote_for_shell(quote_for_system $cur_file);
    $cmd = "$shell /c attrib /q /e -s -h $s";
    warn "$cmd\n";
    $rc = system($cmd);
    die "attrib failed rc = $!(", 0+$!, ")\n" if $rc;
    # system($cmd) or die "attrib failed rc = $!(", 0+$!, ")\n";
    my $s = $/;
    open OUT, ">$cur_file" or die "open failed $!\n";
    binmode OUT;
    undef $/;
    print OUT $buf or die "write failed $!\n";
    close OUT  or die "write failed $!\n";
    $/ = $s;
    warn "Restoring $cur_file attributes and times\n";
    # restore attributes
    chmod $mode, $cur_file;
    # restore times
    utime $atime, $mtime, $cur_file or die "utime failed $!\n";
    print LOG "$cur_file updated\n";
    warn "$cur_file updated\n";
    $files_changed++;
  }

} # while

close IN;
close LOG;
close LIST if $opts{l};

warn "\nFound path references in $files_matched files\n";
warn "Changed path references in $files_changed files\n";

#=== ask_action() Ask for action , return action letter ===

sub ask_action {

  my $action;
  while (1) {
    print 'Change - y)es Y)es to all a) yes to any with \\ A) yes to any n)o N)o to all s)kip rest r)estart R)evert h)elp q)uit ? ';
    $action = <STDIN>;
    chomp $action;
    # print "\n";
    $action = 'n' if $action eq '';
    last if $action =~ /^[YyNnAaRrsq]$/;
print <<EOD if $action eq 'h' || $action eq '?'
  y  Update this path reference in this file
  Y  Update all references to this path in this file
  a  Update all references to any path in this file of form x:\\
  A  Update all references to any path in this file
  n  Do not change this path reference
  N  Do not change any references to this path in this file
  s  Skip rest of file keeping current changes
  r  Revert any changes and restart scan
  R  Revert any changes and skip rest of file
  q  Quit

EOD

  }

  return $action;

} # ask_action

#=== ask_auto_action() Ask for auto action, return action letter ===

sub ask_auto_action {

  my $action;
  while (1) {
    print 'Applied $auto_changes changes - K)eep s)kip R)evert r)estart h)elp q)uit ? ';
    $action = <STDIN>;
    chomp $action;
    # print "\n";
    $action = 'n' if $action eq '';
    last if $action =~ /^[KsRr]$/;
print <<EOD if $action eq 'h' || $action eq '?'
  s  Save changes to this file
  r  Revert any changes and restart scan
  R  Revert any changes and skip rest of file
  q  Quit

EOD
  }

  return $action;

} # ask_auto_action

#=== get_cmd_name() Get script name, set $cmd_name and tag_name ===

sub get_cmd_name {
  $cmd_name = shift;
  $cmd_name = basename($cmd_name);
  $tag_name = $cmd_name;
  $tag_name =~ s/\.[^.]*$//;

} # get_cmd_name

#=== gobble_file(filename) Read file, return buffer ===

sub gobble_file {
  my $path = shift;
  die 'globble_file requires an argument' if not defined $path;
  local *BUF;
  open(BUF, "<$path") or die "open $path $!";
  binmode BUF;
  local $/ = undef;
  my $buf = <BUF>;
  close BUF;

  return $buf;

} # gobble_file

#=== quote_for_system(string) Quote string for system command ===

sub quote_for_system {
  my $s = shift;
  # $s =~ s/\\/\\\\/g;		# Protect \ from expansion
  $s =~ s/\$/\\\$/g;		# Protect $ from expansion
  return $s;

} # quote_for_system

#=== quote_for_shell(string) Quote string for shell ===

sub quote_for_shell {
  my $s = shift;
  # $s =~ s/%/%%/g;		# Protect % from expansion
  $s = '"' . $s . '"' if $s =~ /[ \t]/;	# Quote white
  return $s;

} # quote_for_shell

#=== usage(msg) Print usage message and exit ===

sub usage {
 warn <<EOD;
@_

$usage_msg
EOD
  exit;
}

# The end
